<?php

namespace Drupal\billwerk;

use Drupal\Component\Serialization\Json;
use Drupal\subman\SubmanUtilities;
use Drupal\subman\SubmanWebhookException;
use GuzzleHttp\Client;

/**
 * Class BillwerkApi.
 *
 * Basic functionality to access the remote API of the Billwerk subscription
 * management SaaS.
 *
 * This service class provides the means to actually connect to Billwerk and do
 * thing like:
 * * Authorize against the remote api
 * * Find and retrive remote data
 * * Send requests to the remote api
 * * Handle responses and basic error situations
 *
 * Originally inspired by
 * https://github.com/billwerk/billwerkjs-jquery-php/blob/master/iteroapi.php
 *
 * @todo Refactor re-usable part into an abstract base class in subman.module.
 * @todo Refactor this into a dedicated Drupal-independent composer project?
 */
class BillwerkApi implements BillwerkApiInterface {

  const AUTH_PREFIX = "Authorization: Bearer ";
  const AUTH_CACHE_KEY = 'BillwerkAp::auth_token';
  const AUTH_CACHE_DURATION = 60 * 60;

  /**
   * GuzzleHttp\Client definition.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * Drupal\subman\SubmanUtilities definition.
   *
   * @var \Drupal\subman\SubmanUtilities
   */
  protected $utils;

  /**
   * Constants to be used for assembling URLs.
   */
  /**
   * @todo Convert static properties to class constants.
   */
  static $oAuth                  = "oauth/token";
  static $iteroBase              = "api/v1/";
  static $contractResource       = "contracts/";
  static $customerResource       = "customers/";
  static $customerSearchResource = "Customers";
  static $plansResource          = "Plans";
  static $planVariantsResource   = "PlanVariants";
  const responseCacheName        = 'BillwerkApi::get::requestCache';

  /**
   * Constructs a new BillwerkApi object.
   */
  public function __construct(Client $http_client, SubmanUtilities $subman_utilities) {
    $this->httpClient = $http_client;
    $this->utils = $subman_utilities;
  }

  /**
   * Returns the remote service URL of the SaaS with an optional specific
   * service stub.
   *
   * @param string $service
   *   The additional uri part of the specific service endpoint stub.
   * @param string $system
   *   The additional uri part of the specific service endpoint stub.
   *
   * @return string
   *   The generated remote service url.
   */
  public function serviceUrl(string $service = '', string $system = 'main'): string {
    $url = $this->utils->config('service_url', 'sandbox');
    switch ($url) {
      case 'production':
        $urls = [
          'main' => 'https://app.billwerk.com/',
          'selfservice' => 'https://selfservice.billwerk.com/',
        ];
        break;

      case 'sandbox':
        $urls = [
          'main' => 'https://sandbox.billwerk.com/',
          'selfservice' => 'https://selfservice.sandbox.billwerk.com/',
        ];
        break;

      default:
        throw new \Exception('Unknown service_url "' . $url . '". Allowed values are "production" or "sandbox". Please check your subman.settings!');
    }

    if (!isset($urls[$system])) {
      throw new \Exception('Service URL could not be determined from URL: "' . $url . '", Service: "' . $service . '"' . ', System: "' . $system . '"');
    }

    return $urls[$system] . $service;
  }

  /**
   * Ensures there has been an authentication towards the API.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function ensureAuthorization(): void {
    if (empty($this->getAuthToken())) {
      $client_id = $this->utils->config('client_id');
      $client_secret = $this->utils->config('client_secret');
      $auth = $this->authorize($client_id, $client_secret);
      $this->setAuthToken($auth);
    }
  }

  /**
   * Returns the current token for authentication.
   *
   * @return string|NULL
   *   The current authorization token.
   */
  public function getAuthToken(): ?string {
    // Retrieve cached auth token.
    $auth = $this->utils->loadData(self::AUTH_CACHE_KEY);
    return $auth;
  }

  /**
   * Stores the given token to be used as current authentication token.
   *
   * For not running into quota limits for auth token retrieval,
   * we store a retrieved auth token in cache for some time (1h by default).
   * Compare https://developer.billwerk.io/Docs/ApiIntroduction#authentication
   * By chosing a cache duration that is less than the SaaS own expiration time,
   * we practically circuvent the need for complex detection of invalid-auth
   * responses, re-authentication and re-sending of questes.
   *
   * @param string $auth
   *   The authoriztation token to be set.
   */
  public function setAuthToken(string $auth): void {
    // Cache auth token.
    $this->utils->saveData(self::AUTH_CACHE_KEY, $auth, time() + self::AUTH_CACHE_DURATION);

    // Log retrieval/caching of new auth token (of course only as excerpt, for security reasons).
    // We do this so we can check how often and when auth token have been retrieved.
    $auth_excerpt = substr($auth, strlen(self::AUTH_PREFIX), 5) . '...' . substr($auth, -6);
    $this->utils->log('Authentication token retrieved and cached for @time seconds.', ['auth token excerpt' => $auth_excerpt], ['@time' => self::AUTH_CACHE_DURATION], 'debug');
  }

  /**
   * Retrieves auth token from billwerk API.
   *
   * @param string $clientId
   *   The client ID to be used for authentication.
   * @param string $clientSecret
   *   The client secret to be used for authentication.
   *
   * @return string
   *   The authorization token obtained.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function authorize(string $clientId, string $clientSecret): string {
    // Get oauth token to access remote API.
    $verbose = $this->getErrorStream();
    $curl = curl_init();
    $data = http_build_query(['grant_type' => 'client_credentials']);
    // @todo replace curl with Guzzle.
    curl_setopt_array($curl, [
      CURLOPT_URL => $this->serviceUrl(self::$oAuth),
      CURLOPT_POST => 1,
      CURLOPT_POSTFIELDS => $data,
      CURLOPT_HTTPAUTH => CURLAUTH_BASIC,
      CURLOPT_USERPWD => "$clientId:$clientSecret",
      CURLOPT_RETURNTRANSFER => TRUE,
      CURLOPT_VERBOSE => TRUE,
      CURLOPT_STDERR => $verbose,
      CURLOPT_HTTPHEADER => [
        'Content-Type: application/x-www-form-urlencoded',
      ],
    ]);
    $response = curl_exec($curl);
    $this->exit_on_error($curl, '::authorize');

    // Log (we do not log the full response since it contains the sensitive authentication token).
    $this->utils->log('Authorize', [
      'response size (bytes)' => strlen($response),
    ], [], 'debug');

    // Save auth token.
    $token = json_decode($response);

    return self::AUTH_PREFIX . $token->access_token;
  }

  /**
   * Fetch contract by id from API.
   *
   * @param string $contractId
   *   The id of the contract to get.
   *
   * @return array
   *   The contract data as array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getContract(string $contractId): array {
    return $this->get(self::$contractResource . $contractId);
  }

  /**
   * Fetch customer by id from API.
   *
   * @param string $customerId
   *   The id of the customer to get.
   *
   * @return array
   *   The customer data as array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getCustomer(string $customerId): array {
    return $this->get(self::$customerResource . $customerId);
  }

  /**
   * Fetch customer from API by a search term.
   *
   * @param string $search
   *   The search term to filter customers on.
   *
   * @return array
   *   The customers data as array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function searchCustomers(string $search, int $limit = NULL): array {
    $querystring = "?search=$search";
    if ($limit) {
      $querystring .= '&take=' . $limit;
    }
    return $this->get(self::$customerSearchResource . $querystring);
  }

  /**
   * Update customer in/via API.
   *
   * @param array $customerData
   *   The customer data to update from as array.
   *
   * @return mixed
   *   The result of the operation.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function putCustomer(array $customerData) {
    return $this->put(self::$customerResource . $customerData['Id'], $customerData);
  }

  /**
   * Gets a list of contracts for given customer id.
   *
   * @param string $customerId
   *   The customer id.
   *
   * @return array
   *   The contracts data as nested array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getContractsForCustomer(string $customerId) {
    $url = "customers/$customerId/contracts";
    return $this->get($url);
  }

  /**
   * Gets a list of plans.
   *
   * @return array
   *   The plans data as nested array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getPlans(): array {
    return $this->get(self::$plansResource);
  }

  /**
   * Gets a list of plan variants.
   *
   * @return array
   *   The plan variants data as nested array.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getPlanVariants(): array {
    return $this->get(self::$planVariantsResource);
  }

  /**
   * Retrieves token for accessing hosted self service pages.
   *
   * @param string $contractId
   *   The contract to get the selfservice token for.
   *
   * @return array
   *   The selfservice token information array retrieved.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getSelfserviceTokenArray(string $contractId): array {
    return $this->get(self::$contractResource . $contractId . "/selfServiceToken");
  }

  /**
   * Retrieves the URL for a self-service page for the given contract.
   *
   * @param string $contractId
   *   The contract to get the selfservice url for.
   *
   * @return ?string
   *   The selfservice url retrieved or FALSE.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  public function getUrlSelfservice(string $contractId): ?string {
    $response = $this->getSelfserviceTokenArray($contractId);
    return $response['Url'] ?? FALSE;
  }

  /**
   * Retrieves the URL for a signup page for the plan variant id.
   *
   * @param $planVariantId
   *   The ID of the plan variant to get the signup form url for.
   *
   * @return ?string
   *   The retrieved signup form url or FALSE.
   */
  public function getUrlSignup(string $planVariantId): ?string {
    return $this->serviceUrl("portal/signup.html#/$planVariantId", 'selfservice');
  }

  /**
   * Issue a GET request to the API.
   *
   * Use drupal_static temporary static caching of responses during one Drupal
   * execution, to minimize API calls.
   *
   * @param string $resource
   *   The URL of the resource to get.
   *
   * @return mixed
   *   The retrieved data / result.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  private function get(string $resource) {
    $cache_key = 'GET:' . $resource;
    $cache = $this->utils->config('cache_get', TRUE);
    $responseCache = &drupal_static(self::responseCacheName, []);
    $result = !$cache ? NULL : ($responseCache[$cache_key] ?? NULL);
    $in_cache = !empty($result);

    // If request is not in cache, perform real request.
    if (!$in_cache) {
      // Make sure we're authorized.
      $this->ensureAuthorization();

      // Assemble request options.
      $token = str_replace('Authorization: Bearer ', '', $this->getAuthToken());
      $guzzle_options = [
        'headers' => [
          'Authorization' => 'Bearer ' . $token,
          'Accept' => 'application/json',
          'Content-Type' => 'application/x-www-form-urlencoded',
        ],
        'debug' => FALSE,
        'stream' => FALSE,
      ];

      // Send the request and get the response object.
      $url = $this->serviceUrl(self::$iteroBase) . $resource;
      $response = $this->httpClient->get($url, $guzzle_options);
      $status = $response->getStatusCode();
      // @todo check whether other staus codes have to be considered ok:
      $status_ok = $status == 200;

      if ($status_ok) {
        // Get json string from response.
        $result = $response->getBody()->getContents();

        // Cache response, if caching requests is enabled.
        if ($cache) {
          $responseCache[$cache_key] = $result;
        }
      }

      // Log.
      $this->utils->log(
        'Sending GET request to @url. Status @status "@reason". Result size in bytes: @size',
        [],
        [
          '@url' => $url,
          '@status' => $status,
          '@reason' => $response->getReasonPhrase(),
          '@size' => strlen($result),
        ],
        $status_ok ? 'debug' : 'error'
      );
    }

    // Convert result from json string to PHP structure.
    $result = Json::decode($result);

    return $result;
  }

  /**
   * Resets temporary the response cache.
   */
  public function resetResponseCache(): void {
    $responseCache = &drupal_static(self::responseCacheName, []);
    $responseCache = [];
  }

  /**
   * Issue a PUT request to the API.
   *
   * @param string $resource
   *   The URL of the resource to put to.
   * @param $data
   *   The data to replace the resource with.
   *
   * @return mixed
   *   The result of the operation.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  private function put(string $resource, $data) {
    $this->ensureAuthorization();
    $json = json_encode($data);
    $curl = curl_init();
    curl_setopt_array($curl, [
      CURLOPT_HTTPHEADER => [
        'Content-Type: application/x-www-form-urlencoded',
        'Content-Length: ' . strlen($json),
        $this->getAuthToken(),
      ],
      CURLOPT_POSTFIELDS => $json,
      CURLOPT_CUSTOMREQUEST => "PUT",
    ]);
    $response = $this->request($curl, $resource);
    curl_close($curl);
    return json_decode($response);
  }

  /**
   * Execute API request.
   *
   * @param \CurlHandle $handle
   *   The curl object.
   * @param string $resource
   *   The URL of the resource.
   *
   * @return mixed
   *   The result of the operaton.
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  private function request(\CurlHandle $handle, string $resource) {
    $verbose = $this->getErrorStream();
    curl_setopt_array($handle, [
      CURLOPT_URL => $this->serviceUrl(self::$iteroBase) . $resource,
      CURLOPT_HTTPAUTH => CURLAUTH_ANY,
      CURLOPT_RETURNTRANSFER => TRUE,
      CURLOPT_VERBOSE => TRUE,
      CURLOPT_STDERR => $verbose,
    ]);
    $response = curl_exec($handle);
    $this->exit_on_error($handle, $resource);
    return $response;
  }

  /**
   * Handles exiting in case http response code is not 200.
   *
   * @param \CurlHandle $handle
   * @param string $resource
   *
   * @throws \Drupal\subman\SubmanWebhookException
   */
  private function exit_on_error(\CurlHandle $handle, string $resource) {
    $responseInfo = curl_getinfo($handle);
    $httpResponseCode = $responseInfo['http_code'];

    $error_stream = $this->getErrorStream();
    rewind($error_stream);
    $error_stream_contents = stream_get_contents($error_stream, -1, 0);
    if (function_exists('drush_print')) {
      drush_print($error_stream_contents);
    }

    if ($httpResponseCode != 200) {
      $this->utils->log('Remote service returned error %httpstatuscode on %resource', [
        'responseInfo' => $responseInfo,
        'output stream' => $error_stream_contents,
      ], [
        '%httpstatuscode' => $httpResponseCode,
        '%resource' => $resource,
      ], 'error');
      throw new SubmanWebhookException(SubmanWebhookException::apiError, $responseInfo);
    }
  }

  /**
   * Retrieves an error stream to capute and use error output.
   *
   * @return bool|null|resource
   */
  protected function getErrorStream() {
    // Originally: return fopen('php://stderr', 'w');.
    $stream = &drupal_static('BillwerkApi::getErrorStream', NULL);
    if ($stream == NULL) {
      $stream = fopen('php://memory', 'r+');
      rewind($stream);
    }
    return $stream;
  }

}
